
/*
 * stic - Simple Tiny IRC Client
 *
 * Copyright (C) 2001 by Ian Lynagh <igloo@earth.li>
 * Written by Ian Lynagh <igloo@earth.li>, <igloo@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifdef WIN32
#include <time.h>
#else
#include <pwd.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <termios.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef BUSYBOX
#include "busybox.h" /* xmalloc, xrealloc, xstrdup */
#endif

/* Slower, but smaller */
#define memcpy memmove
#define putchar(c) printf("%c", c)
#define free(ptr) xrealloc(ptr, 0)
#define strchr(s,c) strstr(s, (const char[]){c,0})

/*
   TODO

   Possible features: colour, history, more errors, help
   Tidy up
   Assumes screen width is 80 in a few places (word wrapping, user input)
 */

static char *current_nick; /* What we currently believe our nick to be */

static int out_chars = 0; /* The number of characters of input that are
                             actually on the screen */

/*
   The largest amount of useful text we can input is
   /quote 510_chars\0
   which is length 518. Then we want to be able to get rid of the \0 and
   append \r\n for easy sending, so maximum is 519.
 */
#ifdef BUSYBOX
static char *from_stdin;
#else
static char from_stdin[519];
#endif
/*
   The total number of characters that we have input this line.
   Invariant 1: 0 <= len_from_stdin <= 517
   Corollary 1.1: from_stdin[len_from_stdin] and
                  from_stdin[len_from_stdin + 1] are part of the space
                  allocated for from_stdin
*/
static ssize_t len_from_stdin = 0;

/* struct channel is the building block for the linked list of channels
   we are on (which includes queries). current_channel is a pointer to
   the current active one */
static struct channel {
    char *name;
    struct channel *next;
} *current_channel = NULL;

/* The initial terminal state */
static struct termios termios;

/* The server socket */
static int sock;

/* Init, deinit and main loop functions */
static void usage(char *program_name);
static void set_up_terminal(void);
static void connect_to_server(char * server,
        int port);
static void main_loop(char * nick,
        char * username, char * realname);
static void disconnect_from_server(void);
static void restore_terminal(void);

/* User input handling functions */
static void remove_input(void);
static void print_input(void);
static void remove_channel_and_input(void);
static void print_channel_and_input(void);
static void put_line(char * fmt, ...);
static int deal_user_input(int c);

/* Server IO functions */
static int deal_server_commands(char * from_server,
        int len_from_server);
static void deal_server_command(char *from_server, int len_from_server);
static ssize_t read_data(int fd, char * buff, size_t len);
static int send_to_server(char *buff, ssize_t len);
static char *de_colonise(char * str);
static int end_of_line_offset(char * buff, size_t len);
static int not_end_of_line_offset(char * buff, size_t len);

/* Channel list/current channel maintenance functions */
static void remove_channel (char * name);
static int switch_to_channel (char * new_channel);
static void add_and_switch_to_channel (char * name);

/* Utility functions */
#ifndef BUSYBOX
static void *xmalloc(size_t size);
static void *xrealloc(void *ptr, size_t size);
static char *xstrdup(char *str);
#endif
static void xperror(char * error) __attribute__ ((noreturn));
static char xgetchar (void);
static char *get_time(char *time_start);

int
#ifdef BUSYBOX
    stic_main
#else
    main
#endif
        (int argc, char **argv) {

    char *server = NULL;
    int port = 6667;
    char *nick = NULL;

#ifdef BUSYBOX
    RESERVE_BB_BUFFER(from_stdin_reserve, 519);
    from_stdin = from_stdin_reserve;
#endif
    
    {
        int opt;
        extern char *optarg;

        while ((opt = getopt(argc, argv, "hn:s:p:")) > 0) {
            switch (opt) {
                case 'h':
                    usage(argv[0]);
                    exit(0);
                case 'n':
                    free(nick);
                    nick = xstrdup(optarg);
                    break;
                case 's':
                    free(server);
                    server = xstrdup(optarg);
                    break;
                case 'p':
                    if (sscanf(optarg, "%d", &port) == 1 && port > 0) {
                        break;
                    }
                    /* else fall through */
                default:
                    usage(argv[0]);
                    exit(1);
            }
        }
    }
    if (server == NULL) {
        server = "irc.debian.org";
    }

    if (nick == NULL) {
        uid_t uid;
        struct passwd *pw;

        uid = getuid();
        if ((pw = getpwuid(uid)) == NULL) {
            xperror("getpwuid");
        }
        nick = xstrdup(pw->pw_name);
    }

    set_up_terminal();
    connect_to_server(server, port);
    current_nick = xstrdup("*");
    main_loop(nick, nick, nick);
    putchar('\n');
    disconnect_from_server();
    restore_terminal();
    return 0;
}

/* Init, deinit and main loop functions */

/*
   Print usage information
 */
static void usage(char *program_name) {
    printf("USAGE: %s [-s server] [-p port] [-n nick]\n\n", program_name);
}

/*
   Set the terminal to raw mode, noecho etc and remember the initial
   state in the global variable termios
 */
static void set_up_terminal(void) {
    struct termios tmp;

    if (tcgetattr(0, &termios) == -1) {
        perror("tcgetattr");
        exit(1);
    }
    memcpy(&tmp, &termios, sizeof(struct termios));

    tmp.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
    tmp.c_oflag &= ~OPOST;
    tmp.c_lflag &= ~(ECHO|ICANON|ISIG|IEXTEN);
    tmp.c_cflag &= ~(CSIZE|PARENB);
    tmp.c_cflag |= CS8;

    if (tcsetattr(0, TCSADRAIN, &tmp) == -1) {
        xperror("tcsetattr");
    }
}

/*
   Connect to server server on TCP port port with keepalives enabled.
   The resulting file descriptor is stored in the global int sock.

   server is a NUL terminated string.
   port is the port to connect to.
 */
static void connect_to_server(char * server,
        int port) {
    struct sockaddr_in s_addr;
    struct hostent *he;
    int one = 1;

    sock = socket(AF_INET, SOCK_STREAM, 0);

    memset(&s_addr, 0, sizeof(struct sockaddr_in));
    s_addr.sin_family = AF_INET;
    s_addr.sin_port = htons(port);
    if ((he = gethostbyname(server)) == NULL) {
        xperror("gethostbyname");
    }
    memcpy(&s_addr.sin_addr, he->h_addr, sizeof(he->h_addr));
    setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &one, sizeof one);

    if (connect(sock, (struct sockaddr *)&s_addr, sizeof s_addr) < 0) {
        xperror("connect");
    }
}

/*
   Register with the server and do the select thang

   nick is a NUL terminated string containing the nick we want to use.
   username is a NUL terminated string containing the ident we want to
   use.
   realname is a NUL terminated string containing the real name we want
   to use.
 */
static void main_loop(char * nick, char * username,
        char * realname) {
    char    buff[513];
    ssize_t len_buff = 0;

    len_buff = snprintf(buff, 513, "NICK %s\r\n", nick);
    send_to_server(buff, len_buff);
    /* The server is going to ignore hostname h and servername s anyway,
       so don't bother giving them sensible values */
    /* Could check to see if this is too long */
    len_buff = snprintf(buff, 513, "USER %s h s :%s\r\n", username, realname);
    send_to_server(buff, len_buff);
    len_buff = 0;
    print_channel_and_input();
    fflush(stdout);

    while (1) {
        fd_set read_fds;
        int retval;

        FD_ZERO(&read_fds);
        FD_SET(0, &read_fds);
        FD_SET(sock, &read_fds);

        retval = select(sock + 1, &read_fds, NULL, NULL, NULL);

        if (retval > 0) {
            if (FD_ISSET(0, &read_fds)) {
                if (deal_user_input(xgetchar())) {
                    return;
                }
                fflush(stdout);
            }

            if (FD_ISSET(sock, &read_fds)) {
                retval = read_data(sock, buff + len_buff, 512 - len_buff);
                if (retval == 0) {
                    return;
                }
                len_buff = deal_server_commands(buff, len_buff + retval);
                if (len_buff == 512) {
                    buff[512] = '\0';
                    put_line("Invalid message received: %s", buff);
                    len_buff = 0;
                }
            }
        }
        else if (retval < 0) {
            if (errno != EINTR) {
                xperror("select");
            }
        }
        else {
            /* retval == 0 can't happen */
            xperror("select == 0");
        }
    }
}

/*
   Just close the socket
 */
static void disconnect_from_server(void) {
    close(sock);
}

/*
   Restore the terminal settings to what they were before we started
 */
static void restore_terminal(void) {
    if (tcsetattr(0, TCSADRAIN, &termios) == -1) {
        perror("tcsetattr");
    }
}

/* User input handling functions */

/*
   Delete the characters from the line currently being input from the
   screen (i.e. overwrite them with spaces).
 */
static void remove_input(void) {
    int i;

    for (i = 0; i < out_chars; i++)
        printf("\b \b");
}

/*
   Print the characters from the line currently being input to the
   screen.
 */
static void print_input(void) {
    /* We want to print the last out_chars characters of from_stdin, so
       go to the end (from_stdin + len_from_stdin) and take that many
       off */
    printf("%s", from_stdin + len_from_stdin - out_chars);
}

/*
   Delete the characters from the line currently being input from the
   screen and the initial string indicating the current channel
   (i.e. overwrite them with spaces).
 */
static void remove_channel_and_input(void) {
    int i;

    remove_input();
    i = current_channel == NULL ? 0 : strlen(current_channel->name);
    for (i = i + 3; i; i--)
        printf("\b \b");
}

/*
   Print the current channel prompt and the characters from the line
   currently being input to the screen.
 */
static void print_channel_and_input(void) {
    printf("[%s] ", current_channel == NULL ? "" : current_channel->name);
    print_input();
}

/*
   print a line (first removing the channel and input, then putting them
   back again afterwards). Takes a format string followed by a variable
   number of arguments ala the printf family.
 */
static void put_line(char * fmt, ...) {
    va_list argp;
    int len; /* Length of string to be printed initially, later the
                number of characters printed so far on this line */
    char *line; /* The line of text to be printed */
    char *ptr; /* Pointer to the character we are considering
                  printing */
    int first_word = 1; /* true iff we are printing the first word on a
                           line */
    char *next_word; /* The next word to be printed */
    int len_word = 0; /* The length of the word in next_word so far */

    remove_channel_and_input();

    /* Get the length of the string */
    va_start(argp, fmt);
    len = 1 + vsnprintf(NULL, 0, fmt, argp);
    va_end(argp);

    /* malloc enough space for it and write it */
    line = xmalloc(len);
    va_start(argp, fmt);
    vsnprintf(line, len, fmt, argp);
    va_end(argp);

    len = 0;
    next_word = xmalloc(81);
    next_word[0] = ' ';
    /* Go through it extracting the various irrelevent codes, replace
       anything else unprintable with a '?' and do the word wrapping */
    for(ptr = line; ptr[0] != '\0'; ptr++) {
        switch (ptr[0]) {
            case 2: /* ^B == bold */
            case 18: /* ^R == reverse */
            case 21: /* ^U == underline */
                break;
            case 3: /* ^C == colour */
                if (isdigit(ptr[1])) {
                    ptr++;
                    if (isdigit(ptr[1])) {
                        ptr++;
                    }
                    if (ptr[1] == ',') {
                        ptr++;
                        if (isdigit(ptr[1])) {
                            ptr++;
                        }
                        if (isdigit(ptr[1])) {
                            ptr++;
                        }
                    }
                }
                break;
            default:
                if (!isprint(ptr[0])) {
                    ptr[0] = '?';;
                }
                if (first_word) {
                    len++;
                    if (ptr[0] == ' ') {
                        first_word = 0;
                        len_word = 1;
                    }
                    else {
                        putchar(ptr[0]);
                    }
                }
                else if (ptr[0] == ' ') {
                    len++;
                    next_word[len_word] = '\0';
                    printf("%s", next_word);
                    len_word = 1;
                }
                else {
                    if (len == 79) {
                        next_word[len_word++] = ptr[0];
                        next_word[len_word] = '\0';
                        printf("\r\n%s", next_word + 1);
                        first_word = 1;
                        len = len_word + 1;
                    }
                    else {
                        next_word[len_word++] = ptr[0];
                        len++;
                    }
                }
        }
    }
    if (!first_word) {
        next_word[len_word] = '\0';
        printf("%s", next_word);
    }

    free(line);
    printf("\r\n");
    print_channel_and_input();
    fflush(stdout);
}

/*
   Takes a character c read from stdin and does the appropriate thing
   with it.

   Return true if we should quit (i.e. the user did /quit foo).
*/
static int deal_user_input(int c) {
    char buff[513]; /* Get rid of this and malloc instead? */
    int len_buff;

    switch (c) {
        case '\r':
            remove_input();
            out_chars = 0;

            if (from_stdin[0] != '/' || from_stdin[1] == ' ') {
                int offset;

                offset = from_stdin[0] == '/' ? 2 : 0;

                if (current_channel == NULL) {
                    put_line("You are not on a channel");
                }
                else {
                    len_buff = snprintf(buff, 513,
                            "PRIVMSG %s :%s\r\n", current_channel->name,
                            from_stdin + offset);
                    send_to_server(buff, len_buff);
                    if (    current_channel->name[0] != '#' &&
                            current_channel->name[0] != '&') {
                        put_line("[msg(%s)] %s", current_channel->name,
                                from_stdin + offset);
                    }
                    else {
                        put_line("<%s/%s> %s", current_nick,
                                current_channel->name, from_stdin + offset);
                    }
                }
            }
            else {
                char *params;

                if ((params = strchr(from_stdin, ' ')) == NULL) {
                    if ((strcasecmp(from_stdin, "/QUIT")) == 0) {
                        send_to_server("QUIT Bye\r\n", 18);
                        return 1;
                    }
                    else if ((strcasecmp(from_stdin, "/QUERY")) == 0 ||
                             (strcasecmp(from_stdin, "/PART")) == 0) {
                        if (current_channel != NULL) {
                            if (current_channel->name[0] != '#' &&
                                current_channel->name[0] != '&') {
                                remove_channel(current_channel->name);
                            }
                            else {
                                len_buff = snprintf(buff, 513, "PART %s\r\n",
                                        current_channel->name);
                                send_to_server(buff, len_buff);
                            }
                        }
                    }
                    else {
                        put_line("Unknown command");
                    }
                }
                else {
                    params[0] = '\0';
                    params++;

                    if ((strcasecmp(from_stdin, "/ME")) == 0) {
                        if (current_channel == NULL) {
                            put_line("You are not on a channel");
                        }
                        else {
                            len_buff = snprintf(buff, 513,
                                    "PRIVMSG %s :\001ACTION %s\001\r\n",
                                    current_channel->name, params);
                            send_to_server(buff, len_buff);
                            if (    current_channel->name[0] != '#' &&
                                    current_channel->name[0] != '&') {
                                put_line("* -> %s/%s %s", current_nick,
                                        current_channel->name, params);
                            }
                            else {
                                put_line("* %s/%s %s", current_nick,
                                        current_channel->name, params);
                            }
                        }
                    }
                    else if ((strcasecmp(from_stdin, "/MSG")) == 0) {
                        char *message;

                        if ((message = strchr(params, ' ')) != NULL) {
                            message[0] = '\0';
                            message++;

                            len_buff = snprintf(buff, 513, "PRIVMSG %s :%s\r\n",
                                    params, message);
                            send_to_server(buff, len_buff);
                            if (params[0] != '#' && params[0] != '&') {
                                put_line("[msg(%s)] %s", params, message);
                            }
                            else {
                                put_line("<%s/%s> %s", current_nick,
                                        params, message);
                            }
                        }
                    }
                    else if ((strcasecmp(from_stdin, "/ACTION")) == 0) {
                        char *message;

                        if ((message = strchr(params, ' ')) != NULL) {
                            message[0] = '\0';
                            message++;

                            len_buff = snprintf(buff, 513,
                                    "PRIVMSG %s :\001ACTION %s\001\r\n",
                                    params, message);
                            send_to_server(buff, len_buff);
                            if (params[0] != '#' && params[0] != '&') {
                                put_line("* -> %s/%s %s", current_nick,
                                        params, message);
                            }
                            else {
                                put_line("* %s/%s %s", current_nick,
                                        params, message);
                            }
                        }
                    }
                    else if ((strcasecmp(from_stdin, "/QUIT")) == 0) {
                        memcpy(from_stdin, "QUIT: ", 6);
                        /* Corollary 1.1 */
                        from_stdin[len_from_stdin] = '\r';
                        from_stdin[len_from_stdin + 1] = '\n';
                        send_to_server(from_stdin, len_from_stdin + 2);
                        return 1;
                    }
                    else if ((strcasecmp(from_stdin, "/QUOTE")) == 0) {
                        /* Corollary 1.1 */
                        from_stdin[len_from_stdin] = '\r';
                        from_stdin[len_from_stdin + 1] = '\n';
                        send_to_server(params, len_from_stdin - 5);
                    }
                    else if ((strcasecmp(from_stdin, "/QUERY")) == 0) {
                        if (!switch_to_channel(params)) {
                            add_and_switch_to_channel(params);
                        }
                    }
                    else if ((strcasecmp(from_stdin, "/JOIN")) == 0 ||
                             (strcasecmp(from_stdin, "/PART")) == 0 ||
                             (strcasecmp(from_stdin, "/NICK")) == 0 ||
                             (strcasecmp(from_stdin, "/MODE")) == 0 ||
                             (strcasecmp(from_stdin, "/WHOIS")) == 0 ||
                             (strcasecmp(from_stdin, "/NAMES")) == 0) {
                        /* Could do stricter checking here, but there's not
                           much point, and the server will do anyway */

                        if (!switch_to_channel(params)) {
                            from_stdin[strlen(from_stdin)] = ' ';
                            /* Corollary 1.1 */
                            from_stdin[len_from_stdin] = '\r';
                            from_stdin[len_from_stdin + 1] = '\n';
                            send_to_server(from_stdin + 1,
                                    len_from_stdin + 1);
                        }
                    }
                    else if ((strcasecmp(from_stdin, "/PING")) == 0) {
                        struct timeval tv;

                        gettimeofday(&tv, NULL);
                        len_buff = snprintf(buff, 513,
                                "PRIVMSG %s :\001PING %lu %lu\001\r\n",
                                params, tv.tv_sec, tv.tv_usec);
                        send_to_server(buff, len_buff);
                    }
                    else if ((strcasecmp(from_stdin, "/CTCP")) == 0) {
                        char *message;
                        if ((message = strchr(params, ' ')) != NULL) {
                            message[0] = '\0';
                            len_buff = snprintf(buff, 513,
                                    "PRIVMSG %s :\001%s\001\r\n",
                                    params, message + 1);
                            send_to_server(buff, len_buff);
                        }
                    }
                    else if ((strcasecmp(from_stdin, "/NOTICE") == 0) ||
                             (strcasecmp(from_stdin, "/TOPIC") == 0)) {
                        char *message;
                        if ((message = strchr(params, ' '))
                                != NULL) {
                            from_stdin[strlen(from_stdin)] = ' ';
                            memmove(from_stdin, from_stdin + 1,
                                    1 + message - from_stdin);
                            message[0] = ':';
                            /* Corollary 1.1 */
                            from_stdin[len_from_stdin] = '\r';
                            from_stdin[len_from_stdin + 1] = '\n';
                            send_to_server(from_stdin, len_from_stdin + 1);
                        }
                    }
                    else {
                        put_line("Unknown command");
                    }
                }
            }

            len_from_stdin = 0;
            /* Invariant 1: 0 <= len_from_stdin <= 517 */
            from_stdin[0] = '\0';
        case 127: /* Backspace */
            if (len_from_stdin) {
                /* Invariant 1: 0 <= 0 < len_from_stdin <= 517 */
                int len;
                --len_from_stdin;
                /* Invariant 1: 0 <= len_from_stdin < 517 <= 517 */
                /* Corollary 1.1 */
                from_stdin[len_from_stdin] = '\0';
                /* Relayout the input */
                len = current_channel == NULL
                    ? 0 : strlen(current_channel->name);
                if (out_chars + len + 3 == 60
                        && out_chars < len_from_stdin) {
                    remove_input(); /* Overkill, only need
                                       the \b bits */
                    out_chars += 17;
                    print_input();
                }
                else {
                    printf("\b \b");
                    out_chars--;
                }
            }
            break;
        case 24:
            /* ^X */
            if (current_channel != NULL) {
                remove_channel_and_input();
                current_channel = current_channel->next;
                print_channel_and_input();
            }
            break;
        default: /* Input a character if and only if it's printable */
            /* Invariant 1: 0 <= len_from_stdin <= 517 */
            if (len_from_stdin == 517) {
                return 0;
            }

            /* Invariant 1: 0 <= len_from_stdin < 517 <= 517 */
            if (isprint(c)) {
                int len;

                /* Corollary 1.1 */
                from_stdin[len_from_stdin] = c;
                len_from_stdin++;
                /* Invariant 1: 0 <= 0 < len_from_stdin <= 517 */
                /* Corollary 1.1 */
                from_stdin[len_from_stdin] = '\0';

                len = current_channel == NULL
                    ? 0 : strlen(current_channel->name);
                if (out_chars + len + 3 == 78) {
                    remove_input();
                    out_chars -= 17;
                    print_input();
                }
                else {
                    printf("%c", c);
                    out_chars++;
                }
            }
    }

    return 0;
}

/* Server IO functions */

/*
   deal_server_commands starts at the beginning of the input buffer and
   (logically) executes the following procedure until every character
   remaining in the input buffer is a \r or \n:

     Remove any leading \r and \n characters
     If there is a \r or \n character later on in the buffer then:
       Replace that character with a NUL
       Send the NUL terminated string to deal_server_command
       Remove the NUL terminated string from the buffer

   from_server is a string, not necessarily NUL terminated.
   The first len_from_server bytes are actually set to meaningful
   values.

   Returns the amount of input left and moves the remaining input to the
   beginning of from_server. from_server may not be NUL terminated when
   it returns.
 */
static int deal_server_commands( char * from_server,
        int len_from_server) {
    int offset = 0, i;

    while (offset < len_from_server) {
        if ((i = not_end_of_line_offset(from_server + offset,
                        len_from_server - offset)) == -1) {
            return 0; /* There are only CRs and LFs in the buffer */
        }

        offset += i;

        if ((i = end_of_line_offset(from_server + offset,
                        len_from_server - offset)) == -1) {
            break;
        }

        from_server[offset + i] = '\0';
        deal_server_command(from_server + offset, i);
        offset += i + 1;
    }

#if 0
    if (!offset) {
        return len_from_server;
    }
    else if (offset < len_from_server) {
        memmove(from_server, from_server + offset, len_from_server - offset);
        return len_from_server - offset;
    }
    else {
        return 0;
    }
#endif
    memmove(from_server, from_server + offset, len_from_server - offset);
    return len_from_server - offset;
}

/*
   deal_server_command deals with the a single server command.

   from_server is a null terminated line from the server.
   strlen(from_server) = len_from_server

   During the execution of this function from_server may be changed and
   may not be NULL terminated when it has finished.
 */
static void deal_server_command(char *from_server,
        int len_from_server) {
    char *nick = NULL, *ident = NULL;
    char *orig_params;
    char **params = NULL;
    int num_params = 0;

    {
        int offset = 0;

        if (from_server[0] == ':') {
            while (from_server[offset] != ' ' && from_server[offset] != '\0') {
                if (from_server[offset] == '!') {
                    nick = from_server + 1;
                    from_server[offset] = '\0';
                    ident = nick + offset;
                }
                offset++;
            }
            if (ident == NULL) {
                ident = from_server + 1;
            }

            if (from_server[offset] == ' ') {
                from_server[offset++] = '\0';
            }

            from_server += offset;
        }
    }

    orig_params = xstrdup(from_server);

    {
        int done = 0;
        while (!done) {
            num_params++;
            params = xrealloc(params, num_params * sizeof(char *));

            if (from_server[0] == ':') {
                params[num_params - 1] = xstrdup(from_server + 1);
                done = 1;
            }
            else {
                char *end;

                if ((end = strchr(from_server, ' ')) != NULL) {
                    end[0] = '\0';
                    params[num_params - 1] = xstrdup(from_server);
                    from_server = end + 1;
                }
                else {
                    params[num_params - 1] = xstrdup(from_server);
                    done = 1;
                }
            }
        }
    }

    if (strcasecmp(params[0], "PING") == 0) {
        char buff[513];

        snprintf(buff, 513, "PONG%s\r\n", orig_params + 4);
        send_to_server(buff, len_from_server + 2);
    }
    else if (strcasecmp(params[0], "PRIVMSG") == 0 && num_params == 3) {
        char *tmp;

        if (params[2][0] == 1 &&
                (tmp = strrchr(params[2] + 1, 1)) != NULL &&
                tmp[1] == '\0') {
            char *ctcp_text;

            tmp[0] = '\0';

            if ((ctcp_text = strchr(params[2] + 1, ' ')) == NULL) {
                ctcp_text = "";
            }
            else {
                ctcp_text[0] = '\0';
                ctcp_text++;
            }

            if (strcasecmp(params[2] + 1, "ACTION") == 0) {
                if (strcasecmp(params[1], current_nick) == 0) {
                    put_line("*%s %s", nick, ctcp_text);
                }
                else {
                    put_line("*%s/%s %s", nick, params[1], ctcp_text);
                }
            }
            else {
                int len;
                put_line("%s(%s) requested CTCP %s from %s", nick, ident,
                        params[2] + 1, params[1]);
                if (strcasecmp(params[2] + 1, "PING") == 0) {
                    tmp = xmalloc(513);
                    len = snprintf(tmp, 513,
                            "NOTICE %s :\001PING%s\001\r\n",
                            nick, ctcp_text);
                    send_to_server(tmp, len);
                    free(tmp);
                }
                if (strcasecmp(params[2] + 1, "VERSION") == 0) {
                    tmp = xmalloc(513);
                    len = snprintf(tmp, 513,
                            "NOTICE %s :\001VERSION stic 0.0.1\001\r\n",
                            nick);
                    send_to_server(tmp, len);
                    free(tmp);
                }
            }
        }
        else {
            if (params[1][0] == '#' || params[1][0] == '&') {
                put_line("<%s/%s> %s", nick, params[1], params[2]);
            }
            else {
                put_line("[%s(%s)] %s", nick, ident, params[2]);
            }
        }
    }
    else if (strcasecmp(params[0], "NOTICE") == 0 && num_params == 3) {
        char *tmp;

        if (params[2][0] == 1 &&
                (tmp = strrchr(params[2] + 1, 1)) != NULL &&
                tmp[1] == '\0') {
            char *ctcp_text;
            unsigned long ping_secs;
            long ping_usecs;

            tmp[0] = '\0';

            if ((ctcp_text = strchr(params[2] + 1, ' ')) == NULL) {
                ctcp_text = "";
            }
            else {
                ctcp_text[0] = '\0';
                ctcp_text++;
            }

            if ((strcasecmp(params[2] + 1, "PING")) == 0 &&
                     sscanf(ctcp_text, "%lu %lu",
                        &ping_secs, &ping_usecs) == 2) {
                struct timeval tv;

                gettimeofday(&tv, NULL);
                ping_secs = tv.tv_sec - ping_secs;
                ping_usecs = tv.tv_usec - ping_usecs;
                if (ping_usecs < 0) {
                    ping_usecs += 1000000;
                    ping_secs--;
                }
                put_line("Ping reply from %s: %lu.%02lu second(s)",
                        nick, ping_secs, ping_usecs / 10000);
            }
            else {
                put_line("CTCP %s reply from %s: %s",
                        params[2] + 1, nick, ctcp_text);
            }
        }
        else {
            if (params[1][0] == '#' || params[1][0] == '&') {
                put_line("-%s:%s- %s", nick, params[1], params[2]);
            }
            else {
                put_line("-%s(%s)- %s", nick, ident, params[2]);
            }
        }
    }
    else if (strcasecmp(params[0], "JOIN") == 0 && num_params == 2) {
        if (strcasecmp(current_nick, nick) != 0) {
            put_line("-- %s has joined %s", nick, params[1]);
        }
        else {
            add_and_switch_to_channel(params[1]);
            put_line("You have joined %s", params[1]);
        }
    }
    else if (strcasecmp(params[0], "PART") == 0 && num_params == 2) {
        if (strcasecmp(current_nick, nick) != 0) {
            put_line("-- %s has left %s", nick, params[1]);
        }
        else {
            remove_channel(params[1]);
            put_line("You have left %s", params[1]);
        }
    }
    else if (strcasecmp(params[0], "NICK") == 0 && num_params == 2) {
        if (strcasecmp(nick, current_nick) == 0) {
            free(current_nick);
            current_nick = xstrdup(params[1]);
        }
        put_line("%s changed nick to %s", nick, params[1]);
    }
    else if (strcasecmp(params[0], "TOPIC") == 0 && num_params == 3) {
        put_line("%s set topic on %s to %s", nick, params[1], params[2]);
    }
    else if (strcasecmp(params[0], "MODE") == 0 && num_params >= 2) {
        put_line("%s set mode %s", nick, orig_params + 5);
    }
    else if (strcasecmp(params[0], "QUIT") == 0 && num_params == 2) {
        put_line("SignOff %s (%s)", nick, params[1]);
    }
    else if (strcasecmp(params[0], "KILL") == 0 && num_params == 3) {
        put_line("%s was killed by %s: %s", params[1], nick, params[2]);
    }
    else if (strcasecmp(params[0], "ERROR") == 0 && num_params == 2) {
        put_line("%s", params[1]);
    }
    else if (  isdigit(params[0][0])
            && isdigit(params[0][1])
            && isdigit(params[0][2])
            && params[0][3] == '\0') {

        /* Remove our nick and remember it in current_nick */
        if (num_params > 1) {
            free(current_nick);
            current_nick = xstrdup(params[1]);
        }

        if (strcasecmp(params[0], "332") == 0 && num_params == 4) {
            /* RPL_TOPIC */
            put_line("-- Topic for %s is %s", params[2], params[3]);
        }
        else if (strcasecmp(params[0], "333") == 0 && num_params == 5) {
            /* #lego Igloo 987861645 */
            char *time_str;

            if ((time_str = get_time(params[4])) == NULL) {
                time_str = params[4];
            }
            put_line("-- Topic for %s set by %s at %s",
                    params[2], params[3], time_str);
        }
        else if (strcasecmp(params[0], "353") == 0 && num_params == 5) {
            /* RPL_NAMREPLY */
            /* = #lego :igtest Igloo @Grumpy @Grouchy */
            put_line("-- Users on %s: %s", params[3], params[4]);
        }
        else if (strcasecmp(params[0], "317") == 0 && num_params >= 5) {
            /* RPL_WHOISIDLE */
            /* Igloo 18964 987785098 :seconds idle, signon time */
            unsigned long idle_time;

            if (sscanf(params[3], "%lu", &idle_time) != 1) {
                return;
            }
            put_line("-- %s idle %02lu:%02u:%02u, signon: %s",
                    params[2], idle_time / 3600, (idle_time / 60 % 60),
                    idle_time % 60, get_time(params[4]));
        }
        else {
            /* (num_params > 2) == 1 if there is a space after the nick,
               else 0 */
            from_server = orig_params + 3 + 1
                + strlen(current_nick) + (num_params > 2);

            /* If a parameter begins with a ':' then remove the ':' */
            from_server = de_colonise(from_server);
            /* Print it out */
            put_line("-- %s", from_server);
        }
    }
    else {
        put_line("Got unexpected stuff: %s", orig_params);
    }

    {
        int i;
        for (i = 0; i < num_params; i++) {
            free(params[i]);
        }
        free(params);
        free(orig_params);
    }
}

/*
   read_data is a small wrapper around read(2). It keeps reading until
   the read is not interrupted and then, if there is an error or we
   managed to read too much data, fails. Otherwise it returns the number
   of bytes read ala read(2)

   fd is the file descriptor to which responses are directed if they are
   intended to be given to the server.
   buff is the buffer into which the data should be read.
   len is the maximum length that may be read into the buffer.
 */
static ssize_t read_data(int fd, char * buff,
        size_t len) {
    ssize_t retval;

    /* This while loop can loop forever in theory. That's probably
       bad. */
    do {
        retval = read(fd, buff, len);
    } while (retval < 0 && errno != EINTR);

    if (retval < 0 || retval > len) {
        xperror("read");
    }

    return retval;
}

/*
   send_to_server sends a buffer of data to the server. The buffer
   should have \r\n as the last 2 characters. If the buffer is more than
   512 characters then the 511th and 512th characters will be set to
   \r\n and the length set to 512. It then wraps write(2) in a similar
   way to how read_data wraps read(2) until all the data has been sent.

   buff is the buffer from which the data should be sent. Not
   necessarily NUL terminated.
   len is the length of data that we wish to be sent (but will be set to
   512 if greater than 512).
 */
static int send_to_server(char *buff, ssize_t len) {
    ssize_t retval;

    if (len < 0 || len > 512) {
        buff[510] = '\r';
        buff[511] = '\n';
        len = 512;
    }

    /* This while loop can loop forever in theory. That's probably
       bad. */
    while (len > 0) {
        if ((retval = write(sock, buff, len)) < 0 && errno != EINTR) {
            return -1;
        }

        if (retval > 0) {
            len -= retval;
            buff += retval;
        }
    }
    return 0;
}

/*
   Given a string of arguments, return the string "flattened", so if an
   argument begins with a : then remove the :.

   str is a NUL terminated string.
   A NUL terminated string is returned.
 */
static char *de_colonise(char * str) {
    char *colon;

    if (str[0] == ':') {
        return str + 1;
    }

    if ((colon = strstr(str, " :")) != NULL) {
        memmove(colon + 1, colon + 2, strlen(colon + 1));
    }

    return str;
}

/*
   Return the offset into buff of the first \r or \n if it is in the
   first len characters. Otherwise return -1.

   buff is a character array of length len. It may not be NUL
   terminated.
 */
static int end_of_line_offset(char * buff, size_t len) {
    int i;

    for (i = 0; i < len; i++) {
        if (buff[i] == '\r' || buff[i] == '\n') {
            return i;
        }
    }

    return -1;
}

/*
   Return the offset into buff of the first non-\r and non-\n if it is
   in the first len characters. Otherwise return -1.

   buff is a character array of length len. It may not be NUL
   terminated.
 */
static int not_end_of_line_offset(char * buff, size_t len) {
    int i;

    for (i = 0; i < len; i++) {
        if (buff[i] != '\r' && buff[i] != '\n') {
            return i;
        }
    }

    return -1;
}

/* Channel list/current channel maintenance functions */

/*
   remove_channel takes name, a NUL terminated string, and removes the
   first channel starting from current_channel->next whose name is name
   if such a channel exists. If there are then no channels left then
   current_channel is set to NULL. Otherwise, if current_channel is the
   deleted channel then current_channel is set to the next channel.
 */
static void remove_channel (char * name) {
    struct channel *prev, *channel;

    prev = current_channel;
    channel = prev->next;
    do {
        if (strcasecmp(channel->name, name) == 0) {
            remove_channel_and_input();
            if (channel == prev) {
                current_channel = NULL;
            }
            else {
                prev->next = channel->next;
                if (current_channel == channel)
                    current_channel = current_channel->next;
            }
            free(channel->name);
            free(channel);
            print_channel_and_input();
            break;
        }
        prev = channel;
        channel = channel->next;
    } while (prev != current_channel);
}

/*
   switch_to_channel takes new_channel, a NUL terminated string, and
   sets current_channel to the first channel, starting from
   current_channel, whose name is new_channel. If no such channel exists
   then current_channel is not changed.
 */
static int switch_to_channel (char * new_channel) {
    struct channel *channel;

    if (current_channel != NULL) {
        channel = current_channel;
        do {
            if (strcasecmp(channel->name, new_channel) == 0) {
                remove_channel_and_input();
                current_channel = channel;
                print_channel_and_input();
                return 1;
            }
            channel = channel->next;
        } while (channel != current_channel);
    }

    return 0;
}

/*
   add_and_switch_to_channel takes name, a NUL terminated string, and
   adds a channel with name name after current_channel and then sets
   current_channel to this channel.
 */
static void add_and_switch_to_channel (char * name) {
    struct channel *channel;

    channel = xmalloc(sizeof(struct channel));

    channel->name = xstrdup(name);
    if (current_channel != NULL) {
        channel->next = current_channel->next;
        current_channel->next = channel;
    }
    else {
        channel->next = channel;
    }

    remove_channel_and_input();
    current_channel = channel;
    print_channel_and_input();
}

/* Utility functions */

#ifndef BUSYBOX
/*
   Wrap malloc and handle errors. Actually use realloc for the hard work
   to reduce size
 */
static void *xmalloc(size_t size) {
    void *ptr = NULL;

    if (size) {
        if ((ptr = realloc(ptr, size)) == NULL) {
            xperror("realloc");
        }
        else {
            return ptr;
        }
    }

    return NULL;
}

/*
   Wrap realloc and handle errors.
 */
static void *xrealloc(void *ptr, size_t size) {
    if ((ptr = realloc(ptr, size)) == NULL && size) {
        xperror("realloc");
    }
    else {
        return ptr;
    }
}

/*
   Wrap strdup and handle errors.
 */
static char *xstrdup(char *str) {
    char *str2;

    if ((str2 = strdup(str)) == NULL) {
        xperror("strdup");
    }
    else {
        return str2;
    }
}
#endif

/*
   Restore the terminal, call perror and then exit.
 */
static void xperror(char * error) {
    restore_terminal();
    perror(error);
    exit(1);
}

/*
   Does the same thing as getchar().
 */
static char xgetchar (void) {
    char buff;
    read_data(0, &buff, 1);
    return buff;
}

/*
   Given a NUL terminated string time_start which represents a time_t,
   returna pretty (NUL terminated string) form of the date and time it
   respresents. If time_start does not represent a time_t then return
   the original string.
 */
static char *get_time(char *time_start) {
    time_t secs;
    char *time_str;

    if (sscanf(time_start, "%ld", &secs) == 1) {
        time_str = ctime(&secs);
        time_str[24] = '\0';
        return time_str;
    }

    return time_start;
}

